@@ -200,6 +200,9 @@ ERL_NIF_TERM bitcask_nifs_keydir_itr_next(ErlNifEnv* env, int argc, const ERL_NI
 ERL_NIF_TERM bitcask_nifs_keydir_itr_release(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[]);
 ERL_NIF_TERM bitcask_nifs_keydir_info(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[]);
 ERL_NIF_TERM bitcask_nifs_keydir_release(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[]);
+ERL_NIF_TERM bitcask_nifs_keydir_trim_fstats(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[]);
+
+ERL_NIF_TERM bitcask_nifs_increment_file_id(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[]);
 
 ERL_NIF_TERM bitcask_nifs_create_tmp_file(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[]);
 
@@ -245,6 +248,9 @@ static ErlNifFunc nif_funcs[] =
     {"keydir_itr_release", 1, bitcask_nifs_keydir_itr_release},
     {"keydir_info", 1, bitcask_nifs_keydir_info},
     {"keydir_release", 1, bitcask_nifs_keydir_release},
+    {"keydir_trim_fstats", 2, bitcask_nifs_keydir_trim_fstats},
+
+    {"increment_file_id", 1, bitcask_nifs_increment_file_id},
 
     {"lock_acquire_int",   2, bitcask_nifs_lock_acquire},
     {"lock_release_int",   1, bitcask_nifs_lock_release},
@@ -659,7 +665,7 @@ ERL_NIF_TERM bitcask_nifs_keydir_put_int(ErlNifEnv* env, int argc, const ERL_NIF
             }
             else // file_id is same, change live/total in one entry
             {
-                update_fstats(env, keydir, entry.file_id, 0,
+                update_fstats(env, keydir, entry.file_id, entry.tstamp,
                               0, 1,
                               entry.total_sz - old_entry->total_sz,
                               entry.total_sz);
@@ -1190,6 +1196,71 @@ ERL_NIF_TERM bitcask_nifs_keydir_release(ErlNifEnv* env, int argc, const ERL_NIF
     }
 }
 
+ERL_NIF_TERM bitcask_nifs_increment_file_id(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[])
+{
+    bitcask_keydir_handle* handle;
+
+    if (enif_get_resource(env, argv[0], bitcask_keydir_RESOURCE, (void**)&handle))
+    {
+
+        LOCK(handle->keydir);
+        (handle->keydir->biggest_file_id)++;
+        uint32_t id = handle->keydir->biggest_file_id;
+        UNLOCK(handle->keydir);
+        return enif_make_tuple2(env, ATOM_OK, enif_make_uint(env, id));
+    }
+    else
+    {
+        return enif_make_badarg(env);
+    }
+}
+
+ERL_NIF_TERM bitcask_nifs_keydir_trim_fstats(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[])
+{
+    bitcask_keydir_handle* handle;
+    ERL_NIF_TERM head, tail, list;
+    uint32_t non_existent_entries = 0;
+
+    if (enif_get_resource(env, argv[0], bitcask_keydir_RESOURCE, (void**)&handle)&&
+        enif_is_list(env, argv[1]))
+    {
+        bitcask_keydir* keydir = handle->keydir;
+        
+        LOCK(keydir);
+        uint32_t file_id;
+
+        list = argv[1];
+
+        while (enif_get_list_cell(env, list, &head, &tail))
+        {
+            enif_get_uint(env, head, &file_id);
+
+            khiter_t itr = kh_get(fstats, keydir->fstats, file_id);
+            if (itr != kh_end(keydir->fstats))
+            {
+                bitcask_fstats_entry* curr_f;
+                curr_f = kh_val(keydir->fstats, itr);
+                enif_free_compat(env, curr_f);
+                kh_del(fstats, keydir->fstats, itr);
+            }
+            else
+            {
+                non_existent_entries++;
+            }
+            // if not found, noop, but shouldn't happen.
+            // think about chaning the retval to signal for warning?
+            list = tail;
+        }
+        UNLOCK(keydir);
+        return enif_make_tuple2(env, ATOM_OK, 
+                                enif_make_uint(env, non_existent_entries));
+    }
+    else
+    {
+        return enif_make_badarg(env);
+    }
+}
+
 ERL_NIF_TERM bitcask_nifs_lock_acquire(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[])
 {
     char filename[4096];
