@@ -474,9 +474,9 @@ merge(Dirname, Opts, {FilesToMerge0, ExpiredFiles0}) ->
     %% possible in some circumstances that we'll get an out-of-date
     %% list of files.
     FilesToMerge = [F || F <- FilesToMerge0,
-                         filelib:is_file(F)],
+                         bitcask_fileops:is_file(F)],
     ExpiredFiles = [F || F <- ExpiredFiles0,
-                         filelib:is_file(F)],
+                         bitcask_fileops:is_file(F)],
     merge1(Dirname, Opts, FilesToMerge, ExpiredFiles).
 
 %% Inner merge function, assumes that bitcask is running and all files exist.
@@ -520,7 +520,7 @@ merge1(Dirname, Opts, FilesToMerge, ExpiredFiles) ->
             %% reader/writer comes along in the same VM. Note that we
             %% won't necessarily merge all these files.
             AllFiles = scan_key_files(readable_files(Dirname), LiveKeyDir, [],
-                                      false, true),
+                                      false, true, false),
 
             %% Partition all files to files we'll merge and files we
             %% won't (so that we can close those extra files once
@@ -569,10 +569,6 @@ merge1(Dirname, Opts, FilesToMerge, ExpiredFiles) ->
                                             Acc
                                     end
                             end, {[],[]}, InFiles1),
-    %% Setup our first output merge file and update the merge lock accordingly
-    {ok, Outfile} = bitcask_fileops:create_file(Dirname, Opts),
-    ok = bitcask_lockops:write_activefile(
-           Lock, bitcask_fileops:filename(Outfile)),
 
     %% Initialize the other keydirs we need.
     {ok, DelKeyDir} = bitcask_nifs:keydir_new(),
@@ -582,7 +578,7 @@ merge1(Dirname, Opts, FilesToMerge, ExpiredFiles) ->
                       merge_lock = Lock,
                       max_file_size = get_opt(max_file_size, Opts),
                       input_files = InFiles,
-                      out_file = Outfile,
+                      out_file = fresh,  % will be created when needed
                       merged_files = [],
                       partial = Partial,
                       live_keydir = LiveKeyDir,
@@ -596,8 +592,13 @@ merge1(Dirname, Opts, FilesToMerge, ExpiredFiles) ->
     State1 = merge_files(State),
 
     %% Make sure to close the final output file
-    ok = bitcask_fileops:sync(State1#mstate.out_file),
-    ok = bitcask_fileops:close(State1#mstate.out_file),
+    case State1#mstate.out_file of
+        fresh ->
+            ok;
+        Outfile ->
+            ok = bitcask_fileops:sync(Outfile),
+            ok = bitcask_fileops:close(Outfile)
+    end,
 
     %% Close the original input files, schedule them for deletion,
     %% close keydirs, and release our lock
@@ -606,11 +607,6 @@ merge1(Dirname, Opts, FilesToMerge, ExpiredFiles) ->
     FileNames = [F#filestate.filename || F <- State#mstate.input_files ++ ExpiredFilesFinished],
     [catch set_setuid_bit(F) || F <- FileNames],
     bitcask_merge_delete:defer_delete(Dirname, IterGeneration, FileNames),
-    if InFiles == [] ->
-            bitcask_fileops:delete(Outfile);
-        true ->
-            ok
-    end,
 
     %% Explicitly release our keydirs instead of waiting for GC
     bitcask_nifs:keydir_release(LiveKeyDir),
@@ -638,12 +634,31 @@ needs_merge(Ref) ->
     %% previous merges). Close these files so that we don't leak
     %% file descriptors.
     P = fun(F) ->
-                filelib:is_file(bitcask_fileops:filename(F))
+                bitcask_fileops:is_file(bitcask_fileops:filename(F))
         end,
     {LiveFiles, DeadFiles} = lists:partition(P, State#bc_state.read_files),
 
-    %% Close the dead files
-    [bitcask_fileops:close(F) || F <- DeadFiles],
+    %% Close the dead files and accumulate a list for trimming their 
+    %% fstats entries.
+    DeadIds0 = 
+        [begin
+             bitcask_fileops:close(F),
+             bitcask_fileops:file_tstamp(F)
+         end
+         || F <- DeadFiles],
+    DeadIds = lists:usort(DeadIds0),
+
+    case bitcask_nifs:keydir_trim_fstats(State#bc_state.keydir, 
+                                         DeadIds) of
+        {ok, 0} ->
+            ok;
+        {ok, Warn} ->
+            error_logger:info_msg("Trimmed ~p non-existent fstat entries",
+                                  [Warn]);
+        Err ->
+            error_logger:error_msg("Error trimming fstats entries: ~p",
+                                   [Err])
+    end,
 
     %% Update state with live files
     put_state(Ref, State#bc_state { read_files = LiveFiles }),
@@ -807,7 +822,7 @@ summary_info(Ref) ->
 
     %% Remove any files that don't exist from the initial summary
     Summary = lists:keysort(1, [S || S <- Summary0,
-                                     filelib:is_file(element(2, S))]),
+                                     bitcask_fileops:is_file(element(2, S))]),
     {KeyCount, Summary}.
 
 
@@ -874,31 +889,39 @@ put_state(Ref, State) ->
 reverse_sort(L) ->
     lists:reverse(lists:sort(L)).
 
-scan_key_files([], _KeyDir, Acc, _CloseFile, _EnoentOK) ->
+scan_key_files([], _KeyDir, Acc, _CloseFile, _EnoentOK, _DelZeroFilesP) ->
     Acc;
-scan_key_files([Filename | Rest], KeyDir, Acc, CloseFile, EnoentOK) ->
+scan_key_files([Filename | Rest], KeyDir, Acc, CloseFile, EnoentOK, DelZeroFilesP) ->
     %% Restrictive pattern matching below is intentional
     case bitcask_fileops:open_file(Filename) of
         {ok, File} ->
             FileTstamp = bitcask_fileops:file_tstamp(File),
-            F = fun(K, Tstamp, {Offset, TotalSz}, _) ->
+            F = fun(K, Tstamp, {Offset, TotalSz}, FunAcc) ->
                         bitcask_nifs:keydir_put(KeyDir,
                                                 K,
                                                 FileTstamp,
                                                 TotalSz,
                                                 Offset,
                                                 Tstamp,
-                                                false)
+                                                false),
+                        FunAcc + 1
                 end,
-            bitcask_fileops:fold_keys(File, F, undefined, recovery),
+            case bitcask_fileops:fold_keys(File, F, 0, recovery) of
+                0 when DelZeroFilesP ->
+                    bitcask_fileops:delete(File),
+                    error_logger:info_msg("Removing empty cask file ~p",
+                                          [Filename]);
+                _R ->
+                    ok
+            end,
             if CloseFile == true ->
                     bitcask_fileops:close(File);
                true ->
                     ok
             end,
-            scan_key_files(Rest, KeyDir, [File | Acc], CloseFile, EnoentOK);
+            scan_key_files(Rest, KeyDir, [File | Acc], CloseFile, EnoentOK, DelZeroFilesP);
         {error, enoent} when EnoentOK ->
-            scan_key_files(Rest, KeyDir, Acc, CloseFile, EnoentOK)
+            scan_key_files(Rest, KeyDir, Acc, CloseFile, EnoentOK, DelZeroFilesP)
     end.
 
 %%
@@ -969,7 +992,7 @@ init_keydir_scan_key_files(_Dirname, _Keydir, 0) ->
 init_keydir_scan_key_files(Dirname, KeyDir, Count) ->
     try
         SortedFiles = readable_files(Dirname),
-        _ = scan_key_files(SortedFiles, KeyDir, [], true, false)
+        _ = scan_key_files(SortedFiles, KeyDir, [], true, false, true)
     catch _:_ ->
             init_keydir_scan_key_files(Dirname, KeyDir, Count - 1)
     end.
@@ -1027,8 +1050,6 @@ merge_files(#mstate {  dirname = Dirname,
                        "merge_files: skipping file ~s in ~s: ~p\n",
                        [File#filestate.filename, Dirname, Error]),
                      State
-             after
-                     catch bitcask_fileops:close(File)
              end,
     merge_files(State2#mstate { input_files = Rest }).
 
@@ -1084,14 +1105,26 @@ inner_merge_write(K, V, Tstamp, OldFileId, OldOffset, State) ->
                 %% Start our next file and update state
                 {ok, NewFile} = bitcask_fileops:create_file(
                                   State#mstate.dirname,
-                                  State#mstate.opts),
+                                  State#mstate.opts,
+                                  State#mstate.live_keydir),
                 NewFileName = bitcask_fileops:filename(NewFile),
                 ok = bitcask_lockops:write_activefile(
                        State#mstate.merge_lock,
                        NewFileName),
                 State#mstate { out_file = NewFile };
             ok ->
-                State
+                State;
+            fresh ->
+                %% create the output file and take the lock.
+                {ok, NewFile} = bitcask_fileops:create_file(
+                                  State#mstate.dirname,
+                                  State#mstate.opts,
+                                  State#mstate.live_keydir),
+                NewFileName = bitcask_fileops:filename(NewFile),
+                ok = bitcask_lockops:write_activefile(
+                       State#mstate.merge_lock,
+                       NewFileName),
+                State#mstate { out_file = NewFile }                
         end,
     
     {ok, Outfile, Offset, Size} =
@@ -1192,78 +1225,93 @@ readable_files(Dirname) ->
 do_put(_Key, _Value, State, 0, LastErr) ->
     {{error, LastErr}, State};
 do_put(Key, Value, #bc_state{write_file = WriteFile} = State, Retries, _LastErr) ->
-    case bitcask_fileops:check_write(WriteFile, Key, Value,
-                                     State#bc_state.max_file_size) of
+    {GoForward, State2} =
+      case bitcask_fileops:check_write(WriteFile, Key, Value,
+                                       State#bc_state.max_file_size) of
         wrap ->
             %% Time to start a new write file. Note that we do not close the old
             %% one, just transition it. The thinking is that closing/reopening
             %% for read only access would flush the O/S cache for the file,
             %% which may be undesirable.
-            State2 = wrap_write_file(State);
+            {go_forward, wrap_write_file(State)};
         fresh ->
             %% Time to start our first write file.
             case bitcask_lockops:acquire(write, State#bc_state.dirname) of
                 {ok, WriteLock} ->
-                    {ok, NewWriteFile} = bitcask_fileops:create_file(
-                                           State#bc_state.dirname,
-                                           State#bc_state.opts),
-                    ok = bitcask_lockops:write_activefile(
-                           WriteLock,
-                           bitcask_fileops:filename(NewWriteFile)),
-                    State2 = State#bc_state{ write_file = NewWriteFile,
-                                             write_lock = WriteLock };
+                    case bitcask_fileops:create_file(
+                           State#bc_state.dirname,
+                           State#bc_state.opts,
+                           State#bc_state.keydir) of
+                        {ok, NewWriteFile} ->
+                            ok = bitcask_lockops:write_activefile(
+                                   WriteLock,
+                                   bitcask_fileops:filename(NewWriteFile)),
+                            {go_forward, State#bc_state{ write_file = NewWriteFile,
+                                                         write_lock = WriteLock }};
+                        {error, Error} ->
+                            error_logger:error_msg("~s:create_file() failed with ~p, retrying put, write_file=~p\n",
+                                                   [?MODULE, Error, State#bc_state.write_file]),
+                            {Error, wrap_write_file(State#bc_state{write_lock=WriteLock})}
+                    end;
                 {error, Reason} ->
-                    State2 = undefined,
                     throw({error, {write_locked, Reason, State#bc_state.dirname}})
             end;
-
         ok ->
-            State2 = State
+            {go_forward, State}
     end,
 
-    Tstamp = bitcask_time:tstamp(),
-    {ok, WriteFile2, Offset, Size} = bitcask_fileops:write(
-                                       State2#bc_state.write_file,
-                                       Key, Value, Tstamp),
-    case bitcask_nifs:keydir_put(State2#bc_state.keydir, Key,
-                                 bitcask_fileops:file_tstamp(WriteFile2),
-                                 Size, Offset, Tstamp, true) of
-        ok ->
-            {ok, State2#bc_state { write_file = WriteFile2 }};
-        already_exists ->
-            %% Assuming the timestamps in the keydir are
-            %% valid, there is an edge case where the merge thread
-            %% could have rewritten this Key to a file with a greater
-            %% file_id. Rather than synchronize the merge/writer processes, 
-            %% wrap to a new file with a greater file_id and rewrite
-            %% the key there.  Limit the number of recursions in case
-            %% there is a different issue with the keydir.
-            State3 = wrap_write_file(State2#bc_state { write_file = WriteFile2 }),
-            do_put(Key, Value, State3, Retries - 1, already_exists)
+    if GoForward =/= go_forward ->
+            do_put(Key, Value, State2, Retries - 1, GoForward);
+       true ->
+            Tstamp = bitcask_time:tstamp(),
+            {ok, WriteFile2, Offset, Size} = bitcask_fileops:write(
+                                               State2#bc_state.write_file,
+                                               Key, Value, Tstamp),
+            case bitcask_nifs:keydir_put(State2#bc_state.keydir, Key,
+                                         bitcask_fileops:file_tstamp(WriteFile2),
+                                         Size, Offset, Tstamp, true) of
+                ok ->
+                    {ok, State2#bc_state { write_file = WriteFile2 }};
+                already_exists ->
+                    %% Assuming the timestamps in the keydir are
+                    %% valid, there is an edge case where the merge thread
+                    %% could have rewritten this Key to a file with a greater
+                    %% file_id. Rather than synchronize the merge/writer processes,
+                    %% wrap to a new file with a greater file_id and rewrite
+                    %% the key there.  Limit the number of recursions in case
+                    %% there is a different issue with the keydir.
+                    State3 = wrap_write_file(State2#bc_state { write_file = WriteFile2 }),
+                    do_put(Key, Value, State3, Retries - 1, already_exists)
+            end
     end.
 
-
-wrap_write_file(#bc_state{write_file = WriteFile} = State) ->
-    LastWriteFile = bitcask_fileops:close_for_writing(WriteFile),
+wrap_write_file(#bc_state{write_file = WriteFile,
+                          read_files = ReadFiles} = State) ->
+    NewReadFiles = case bitcask_fileops:close_for_writing(WriteFile) of
+                       ok ->
+                           ReadFiles;
+                       LastWriteFile ->
+                           [LastWriteFile|ReadFiles]
+                   end,
     {ok, NewWriteFile} = bitcask_fileops:create_file(
                            State#bc_state.dirname,
-                           State#bc_state.opts),
+                           State#bc_state.opts,
+                           State#bc_state.keydir),
     ok = bitcask_lockops:write_activefile(
            State#bc_state.write_lock,
            bitcask_fileops:filename(NewWriteFile)),
     State#bc_state{ write_file = NewWriteFile,
-                    read_files = [LastWriteFile | 
-                                  State#bc_state.read_files]}.
+                    read_files = NewReadFiles}.
 
 set_setuid_bit(File) ->
     %% We're intentionally opinionated about pattern matching here.
-    {ok, FI} = file:read_file_info(File),
+    {ok, FI} = bitcask_fileops:read_file_info(File),
     NewFI = FI#file_info{mode = FI#file_info.mode bor 8#4000},
-    ok = file:write_file_info(File, NewFI).
+    ok = bitcask_fileops:write_file_info(File, NewFI).
 
 has_setuid_bit(File) ->
     try
-        {ok, FI} = file:read_file_info(File),
+        {ok, FI} = bitcask_fileops:read_file_info(File),
         FI#file_info.mode band 8#4000 == 8#4000
     catch _:_ ->
             false
@@ -1359,17 +1407,11 @@ default_dataset() ->
 %% HACK: Terrible hack to ensure that the .app file for
 %% bitcask is available on the code path. Assumption here
 %% is that we're running in .eunit/ as part of rebar.
-%% MORE HACK: init the io mode (nif or erlang) here.
 a0_test() ->
     code:add_pathz("../ebin"),
     application:start(erlang),
-    Mode0 = case os:getenv("IOMODE") of
-                false -> "erlang";
-                Else -> Else
-            end,
-    Mode = list_to_atom(Mode0),
-    application:set_env(bitcask, io_mode, Mode),
-    error_logger:info_msg("Set IO mode to: ~p", [Mode]).
+    Mode = bitcask_io:determine_file_module(),
+    error_logger:info_msg("Bitcask IO mode is: ~p\n", [Mode]).
 
 roundtrip_test() ->
     os:cmd("rm -rf /tmp/bc.test.roundtrip"),
@@ -1650,9 +1692,9 @@ expire_merge_test() ->
     timer:sleep(1100),
     ok = merge("/tmp/bc.test.mergeexpire",[{expiry_secs,1}]),
 
-    %% Verify we've now only got one file
+    %% With lazy merge file creation there will be no files.
     ok = bitcask_merge_delete:testonly__delete_trigger(),
-    1 = length(readable_files("/tmp/bc.test.mergeexpire")),
+    0 = length(readable_files("/tmp/bc.test.mergeexpire")),
 
     %% Make sure all the data is present
     B = bitcask:open("/tmp/bc.test.mergeexpire"),
@@ -1663,6 +1705,36 @@ expire_merge_test() ->
     close(B),
     ok.
 
+single_key_expire_test_() ->
+    {timeout, 60, fun single_key_expire_body/0}.
+
+% At a customer site, having the same key written to a single file
+% was observed to not update the newest/latest tstamp stats and result
+% in the file considered expired before it was due.
+% The reason is the stats are not updated correctly when putting in the
+% same file. The develop branch has a test only for the stats, but this is
+% the original test that helped narrow down the problem.
+single_key_expire_body() ->
+    Dir = "/tmp/bc.single.key.expire",
+    K = <<"key">>,
+    Data = [{K, <<0:128/integer-unit:8>>} || _ <- lists:seq(1,4)],
+    B1 = init_dataset(Dir, [{max_file_size, 1024}, {expiry_secs, 4}], Data),
+    timer:sleep(6000),
+    ok = bitcask:put(B1, K, <<"val1">>),
+    timer:sleep(1100),
+    ?assertEqual([K], bitcask:list_keys(B1)),
+    ok = bitcask:close(B1),
+    B2 = bitcask:open(Dir, [read_write, {expiry_secs, 4}]),
+    case needs_merge(B2) of
+        {true, MFiles} ->
+            merge(Dir, [{expiry_secs, 4}], MFiles);
+        false ->
+            ok
+    end,
+    ?assertEqual([K], bitcask:list_keys(B2)),
+    bitcask:close(B2).
+
+
 fold_deleted_test() ->    
     os:cmd("rm -rf /tmp/bc.test.fold_delete"),
     B = bitcask:open("/tmp/bc.test.fold_delete",
@@ -1976,6 +2048,50 @@ truncated_merge_test() ->
                         {KV, {ok, V}} = {KV, bitcask:get(B, K)}
                 end, undefined, GoodData).
 
+gh137_regression_test_() ->
+    {timeout, 300,
+     fun() ->
+             %% Case 1-2: there are ok cask files/hints following N
+             %% Case 3: mangle either a data file or hint file
+             %% Case 4-5: a mangled data/hint file exists but not the
+             %%           corresponding type
+             [{ok, N} = {gh137_regression_test(N, Blob, Type), N} ||
+                 N <- [1,2,3,4,5],
+                 Blob <- [<<>>,
+                          <<"foobar">>,
+                          <<N:(80*8)>> % 80 bytes
+                         ],
+                 Type <- ["data", "hint"]],
+             ok
+     end}.
+
+gh137_regression_test(N, Blob, Type) ->
+    Dir = "/tmp/bc.test.gh137",
+    os:cmd("rm -rf " ++ Dir),
+    os:cmd("mkdir " ++ Dir),
+    try
+        close(init_dataset(Dir, [{max_file_size, 1}], default_dataset())),
+
+        %% Verify number of files in directory
+        3 = length(readable_files(Dir)),
+
+        %% GH 137: make a 0-byte data file
+        file:write_file(Dir ++ "/" ++
+                        integer_to_list(N) ++ ".bitcask." ++ Type, Blob),
+
+        B = bitcask:open(Dir, [read_write]),
+        KsVs = [{<<"k10">>, <<"v10a">>},
+                {<<"k11">>, <<"v11a">>},
+                {<<"k12">>, <<"v12a">>},
+                {<<"k13">>, <<"v13a">>}],
+        [ok = bitcask:put(B, K, V) || {K, V} <- KsVs],
+        [{ok, V} = bitcask:get(B, K) || {K, V} <- KsVs],
+        bitcask:close(B),
+        ok
+    after
+        os:cmd("rm -rf " ++ Dir)
+    end.
+
 truncate_file(Path, Offset) ->
     {ok, FH} = file:open(Path, [read, write]),
     {ok, Offset} = file:position(FH, Offset),
