@@ -38,6 +38,8 @@
          keydir_wait_pending/1,
          keydir_info/1,
          keydir_release/1,
+         increment_file_id/1,
+         keydir_trim_fstats/2,
          lock_acquire/2,
          lock_release/1,
          lock_readdata/1,
@@ -118,6 +120,8 @@
          {integer(), integer(), boolean()}}.
 -spec keydir_release(reference()) ->
         ok.
+-spec keydir_trim_fstats(reference(), [integer()]) ->
+        {ok, integer()} | {error, atom()}.
 -spec lock_acquire(string(), integer()) ->
         {ok, reference()} | {error, atom()}.
 -spec lock_release(reference()) ->
@@ -238,6 +242,10 @@ keydir_itr_next_int(_Ref) ->
 keydir_itr_release(_Ref) ->
     ok.
 
+
+increment_file_id(_Ref) ->
+    erlang:nif_error({error, not_loaded}).
+
 keydir_fold(Ref, Fun, Acc0, MaxAge, MaxPuts) ->
     FrozenFun = fun() -> keydir_fold_cont(keydir_itr_next(Ref), Ref, Fun, Acc0) end,
     keydir_frozen(Ref, FrozenFun, MaxAge, MaxPuts).
@@ -305,6 +313,9 @@ keydir_info(_Ref) ->
 keydir_release(_Ref) ->
     erlang:nif_error({error, not_loaded}).
 
+keydir_trim_fstats(_Ref, _IDList) ->
+    erlang:nif_error({error, not_loaded}).
+
 
 lock_acquire(Filename, IsWriteLock) ->
     bitcask_bump:big(),
